"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("tocca");
var ko = require("knockout");
var style_1 = require("./style");
var eventListener_1 = require("./eventListener");
var userAgent_1 = require("./userAgent");
var uiElementPropertyName = "uiElement";
var mouseEvents = {
    contextmenu: true,
    mousedown: true,
    mouseenter: true,
    mouseleave: true,
    mousemove: true,
    mouseout: true,
    mouseover: true,
    mouseup: true
};
var touchEvents = {
    touchcancel: true,
    touchend: true,
    touchmove: true,
    touchstart: true,
    longtap: true,
    swipeleft: true,
    swiperight: true,
    swipeup: true,
    swipedown: true
};
function isGlobalEvent(event) {
    return !(mouseEvents[event] || touchEvents[event]);
}
exports.settings = {
    tapDelay: 300
};
var observerOptions = {
    childList: true,
    subtree: true
};
var mutationObserver = new MutationObserver(function (mutationRecords) {
    mutationRecords.forEach(function (mutationRecord) {
        if (mutationRecord.removedNodes) {
            mutationRecord.removedNodes.forEach(function (removedNode) {
                var uiElement = removedNode[uiElementPropertyName];
                if (uiElement && uiElement.disposeOnDomRemoval && uiElement.dispose) {
                    uiElement.dispose();
                }
            });
        }
    });
});
mutationObserver.observe(document.body, observerOptions);
function setObserverOptions(options) {
    observerOptions = options;
}
exports.setObserverOptions = setObserverOptions;
function isNotNullOrUndefined(value) {
    return !(value === null || typeof value === "undefined");
}
var UiElement = (function () {
    function UiElement(tagName, elementType, options) {
        var _this = this;
        this._attributes = {};
        this._classes = {};
        this._style = {};
        this._userSelectable = true;
        this._addControlToDataDictionary = false;
        this._children = new Array();
        this._knockoutSubscriptions = Array();
        this._styleDisplayVisible = "";
        this._visible = true;
        this.diposeOnDomRemoval = false;
        this.dispose = function () {
            _this.remove();
            if (_this._children) {
                _this._children.forEach(function (child) {
                    child.dispose();
                });
            }
            _this._knockoutSubscriptions.forEach(function (subscription) {
                subscription.dispose();
            });
        };
        this._tagName = tagName;
        this._elementType = elementType || "UiElement";
        this.options = options || {};
        if (options) {
            if (isNotNullOrUndefined(options.id)) {
                this.setAttribute("id", options.id);
            }
            if (options.elementType) {
                this._elementType = ko.unwrap(options.elementType);
                if (ko.isObservable(options.elementType)) {
                    this.addSubscription(options.elementType, function () {
                        _this.setElementType(elementType);
                    });
                }
            }
            if (options.classNames) {
                this.setClasses(ko.unwrap(options.classNames), true);
                if (ko.isObservable(options.classNames)) {
                    this.addSubscription(options.classNames, function (classNames) {
                        _this.setClasses(classNames, true);
                    });
                }
            }
            if (options.attributes) {
                this._setAttributes(ko.unwrap(options.attributes));
                if (ko.isObservable(options.attributes)) {
                    this.addSubscription(options.attributes, function (attributes) {
                        _this._setAttributes(attributes);
                    });
                }
            }
            if (options.eventListeners) {
                this._eventListeners = options.eventListeners;
            }
            if (options.style) {
                var _loop_1 = function (styleProperty) {
                    var styleValue = options.style[styleProperty];
                    this_1._style[styleProperty] = ko.unwrap(styleValue);
                    if (ko.isObservable(styleValue)) {
                        this_1.addSubscription(styleValue, function (newValue) {
                            var _a;
                            _this.setStyle((_a = {}, _a[styleProperty] = newValue, _a));
                        });
                    }
                };
                var this_1 = this;
                for (var _i = 0, _a = Object.keys(options.style); _i < _a.length; _i++) {
                    var styleProperty = _a[_i];
                    _loop_1(styleProperty);
                }
            }
            if (options.userSelectable === false) {
                this._userSelectable = false;
            }
            if (isNotNullOrUndefined(options.innerHtml)) {
                this.setInnerHtml(ko.unwrap(options.innerHtml));
                if (ko.isObservable(options.innerHtml)) {
                    this.addSubscription(options.innerHtml, function (innerHtml) {
                        _this.setInnerHtml(innerHtml);
                    });
                }
            }
            if (isNotNullOrUndefined(options.innerText)) {
                this.setInnerText(ko.unwrap(options.innerText));
                if (ko.isObservable(options.innerText)) {
                    this.addSubscription(options.innerText, function (newValue) {
                        _this.setInnerText(newValue);
                    });
                }
            }
            if (options.addControlToDataDictionary) {
                this._addControlToDataDictionary = true;
            }
            if (options.children) {
                this._children = ko.unwrap(options.children);
                if (ko.isObservable(options.children)) {
                    this.addSubscription(options.children, function (newChildren) {
                        _this.setChildren(newChildren);
                    });
                }
            }
            if (options.dispose) {
                var dispose = this.dispose;
                this.dispose = function () {
                    dispose();
                    options.dispose();
                };
            }
            if (options.refresh) {
                this.refresh = options.refresh;
            }
            if (options.visible) {
                var visible = ko.unwrap(options.visible);
                if (visible === false) {
                    this._visible = false;
                }
                if (ko.isObservable(options.visible)) {
                    this.addSubscription(options.visible, function (newValue) {
                        _this.setVisibility(newValue);
                    });
                }
            }
        }
    }
    UiElement.prototype._addEventListenerToElement = function (type, listener, options) {
        if (!type) {
            throw new Error("type is mandatory");
        }
        if (!listener) {
            throw new Error("listener is mandatory");
        }
        if (!options) {
            options = {};
        }
        if (!options.passive) {
            options.passive = false;
        }
        if (!options.once) {
            options.once = false;
        }
        if (this.element.addEventListener) {
            this.element.addEventListener(type, function (event) {
                listener(event);
            }, {
                passive: options.passive,
                once: options.once
            });
        }
        else if (this.element.attachEvent) {
            this.element.attachEvent(type, function (event) {
                listener(event);
            });
        }
        else {
            throw "Your browser does not support 'addEventListener'";
        }
    };
    UiElement.prototype._setAttributes = function (attributes) {
        var _this = this;
        var existingAttributeNames = Object.keys(this._attributes);
        this._attributes = {};
        attributes.forEach(function (attribute) {
            _this._attributes[attribute.name] = ko.unwrap(attribute.value);
            if (ko.isObservable(attribute.value)) {
                _this.addSubscription(attribute.value, function (newValue) {
                    _this.setAttribute(attribute.name, newValue);
                });
            }
        });
        if (existingAttributeNames) {
            existingAttributeNames.forEach(function (existingAttributeName) {
                if (!_this._attributes[existingAttributeName]) {
                    _this.removeAttribute(existingAttributeName);
                }
            });
        }
    };
    UiElement.prototype.build = function () {
        var _this = this;
        if (this._elementType) {
            this.setAttribute("elementtype", this._elementType);
        }
        var display = this.getStyleValue("display");
        if (display && display !== "none") {
            this._styleDisplayVisible = display;
        }
        this.setVisibility(this._visible);
        ko.applyBindingsToNode(this.element, {
            style: this._style,
            attr: this._attributes
        });
        if (isNotNullOrUndefined(this._innerHtml)) {
            this.element.innerHTML = this._innerHtml;
        }
        if (isNotNullOrUndefined(this._innerText)) {
            this.element.innerText = this._innerText;
        }
        this.element.className = Object.keys(this._classes).join(" ");
        if (this._userSelectable === false) {
            this._addEventListenerToElement("selectstart", function () { return false; }, { passive: true });
            this.element.style.userSelect = "none";
        }
        if (this._children) {
            this._children.forEach(function (child) {
                _this.element.appendChild(child.render());
            });
        }
        if (this._addControlToDataDictionary) {
            if (!this.element.data) {
                this.element.data = {};
            }
            this.element.data.uiElement = this;
        }
    };
    UiElement.prototype.render = function () {
        var _this = this;
        if (this.element) {
            this.remove();
        }
        if (typeof this.build === "function") {
            this.element = document.createElement(this._tagName);
            this.build();
            var clickEventListeners_1 = new Array();
            var doubleClickEventListeners_1 = new Array();
            if (this._eventListeners) {
                var eventListeners = this._eventListeners;
                eventListeners.forEach(function (eventListener) {
                    if (eventListener.type === "click") {
                        clickEventListeners_1.push(eventListener);
                        return;
                    }
                    if (eventListener.type === "dblclick") {
                        doubleClickEventListeners_1.push(eventListener);
                        return;
                    }
                    if (eventListener.type === "dragstart") {
                        _this.setAttribute("draggable", "true");
                    }
                    if (isGlobalEvent(eventListener.type)) {
                        _this._addEventListenerToElement(eventListener.type, function (event) {
                            if (eventListener.options) {
                                var options = eventListener.options;
                                var keyEvent = event;
                                if (options.altKey && !keyEvent.altKey) {
                                    return;
                                }
                                if (options.shiftKey && !keyEvent.shiftKey) {
                                    return;
                                }
                                if (options.ctrlKey && !keyEvent.ctrlKey) {
                                    return;
                                }
                                if (options.eventKey && options.eventKey !== keyEvent.key) {
                                    return;
                                }
                            }
                            eventListener.listener.call(_this, event);
                        }, eventListener.options);
                    }
                    else if (mouseEvents[eventListener.type] && userAgent_1.userAgent.device.supportsMouseEvents) {
                        _this._addEventListenerToElement(eventListener.type, eventListener.listener, eventListener.options);
                    }
                    else if (touchEvents[eventListener.type] && userAgent_1.userAgent.device.supportsTouchEvents) {
                        _this._addEventListenerToElement(eventListener.type, eventListener.listener, eventListener.options);
                    }
                });
                clickEventListeners_1.forEach(function (clickEventListener) {
                    if (userAgent_1.userAgent.device.supportsTouchEvents) {
                        _this._addEventListenerToElement("tap", clickEventListener.listener, clickEventListener.options);
                    }
                    if (userAgent_1.userAgent.device.supportsMouseEvents) {
                        _this._addEventListenerToElement(clickEventListener.type, clickEventListener.listener, clickEventListener.options);
                    }
                });
                doubleClickEventListeners_1.forEach(function (doubleClickEventListener) {
                    if (userAgent_1.userAgent.device.supportsTouchEvents) {
                        _this._addEventListenerToElement("dbltap", doubleClickEventListener.listener, doubleClickEventListener.options);
                    }
                    if (userAgent_1.userAgent.device.supportsMouseEvents) {
                        _this._addEventListenerToElement(doubleClickEventListener.type, doubleClickEventListener.listener, doubleClickEventListener.options);
                    }
                });
            }
            return this.element;
        }
        throw "The build method of this UiElement has not been defined";
    };
    UiElement.prototype.refresh = function () { };
    UiElement.prototype.remove = function () {
        if (this.element) {
            if (this.element.remove) {
                this.element.remove();
            }
            else {
                if (this.element.parentElement) {
                    this.element.parentElement.removeChild(this.element);
                }
            }
            this.element = null;
        }
    };
    UiElement.prototype.handleMessage = function (message) {
        console.log(JSON.stringify(message));
    };
    UiElement.prototype.getElement = function () {
        return this.element;
    };
    UiElement.prototype.focus = function (options) {
        if (this.element) {
            this.element.focus(options);
        }
    };
    UiElement.prototype.focusPreviousSibling = function (options) {
        if (this.element && this.element.previousElementSibling && this.element.nextSibling instanceof HTMLElement) {
            this.element.nextSibling.focus(options);
        }
    };
    UiElement.prototype.focusNextSibling = function (options) {
        if (this.element && this.element.previousElementSibling && this.element.nextSibling instanceof HTMLElement) {
            this.element.nextSibling.focus(options);
        }
    };
    UiElement.prototype.setElementType = function (elementType) {
        this._elementType = elementType;
        this.setAttribute("elementType", elementType);
    };
    UiElement.prototype.setClasses = function (classNames, replace) {
        var _this = this;
        if (replace) {
            this._classes = {};
        }
        if (classNames) {
            classNames.forEach(function (className) {
                _this._classes[className] = true;
            });
        }
        if (this.element) {
            this.element.className = Object.keys(this._classes).join(" ");
        }
    };
    UiElement.prototype.removeClasses = function (classNames) {
        var _this = this;
        if (classNames) {
            classNames.forEach(function (className) {
                if (_this._classes[className]) {
                    delete _this._classes[className];
                }
                if (_this.element) {
                    _this.element.classList.remove(className);
                }
            });
        }
    };
    UiElement.prototype.setInnerHtml = function (innerHtml) {
        this._innerHtml = innerHtml;
        if (this.element) {
            this.element.innerHTML = innerHtml;
        }
    };
    UiElement.prototype.setInnerText = function (innerText) {
        this._innerText = innerText;
        if (this.element) {
            this.element.innerText = innerText;
        }
    };
    UiElement.prototype.setAttribute = function (name, value) {
        this._attributes[name] = value;
        if (this.element) {
            this.element.setAttribute(name, value);
        }
    };
    UiElement.prototype.removeAttribute = function (attributeName) {
        if (this._attributes[attributeName]) {
            delete this._attributes[attributeName];
        }
        if (this.element) {
            this.element.removeAttribute(attributeName);
        }
    };
    UiElement.prototype.getAttribute = function (attributeName) {
        if (this.element) {
            return this.element.getAttribute(attributeName);
        }
        else {
            return this._attributes[attributeName];
        }
    };
    UiElement.prototype.getStyle = function () {
        var _this = this;
        var cssProperties = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            cssProperties[_i] = arguments[_i];
        }
        if (!cssProperties || cssProperties.length === 0) {
            return this._style;
        }
        var style = {};
        if (cssProperties) {
            cssProperties.forEach(function (cssProperty) {
                if (_this._style[cssProperty]) {
                    style[cssProperty] = _this._style[cssProperty];
                }
                else if (_this.element) {
                    style[cssProperty] = _this.element.style[cssProperty];
                }
            });
        }
        return style;
    };
    UiElement.prototype.getStyleValue = function (cssProperty) {
        if (this._style[cssProperty]) {
            return this._style[cssProperty];
        }
        else if (this.element) {
            return this.element.style[cssProperty];
        }
        return null;
    };
    UiElement.prototype.setStyle = function (newStyle, overwriteExisting) {
        var _this = this;
        if (newStyle) {
            Object.keys(newStyle).forEach(function (key) {
                if (_this._style.hasOwnProperty(key)) {
                    if (!overwriteExisting) {
                        return;
                    }
                }
                var newValue = newStyle[key];
                _this._style[key] = newValue;
                if (_this.element) {
                    if (newValue === null || newValue === undefined) {
                        if (userAgent_1.userAgent.browser.isInternetExplorer) {
                            _this.element.style[key] = style_1.defaultStyle[key];
                        }
                        else {
                            _this.element.style[key] = null;
                        }
                    }
                    else {
                        _this.element.style[key] = newValue;
                    }
                }
            });
        }
    };
    UiElement.prototype.setStyleNonStandard = function (name, value, overwriteExisting) {
        var _a;
        this.setStyle((_a = {}, _a[name] = value, _a), overwriteExisting);
    };
    UiElement.prototype.deleteStyle = function (style) {
        var _this = this;
        var styles;
        if (style instanceof Array) {
            styles = style;
        }
        else if (style instanceof Object) {
            styles = Object.keys(style);
        }
        else {
            styles = [style];
        }
        styles.forEach(function (cssProperty) {
            if (_this._style && _this._style.hasOwnProperty(cssProperty)) {
                delete _this._style[cssProperty];
            }
            if (_this.element) {
                _this.element.style[cssProperty] = undefined;
            }
        });
    };
    UiElement.prototype.setVisibility = function (visible) {
        this._visible = visible;
        if (this.element) {
            if (visible) {
                this.element.style.display = this._styleDisplayVisible;
            }
            else {
                this.element.style.display = "none";
            }
        }
    };
    UiElement.prototype.addSubscription = function (observable, callback) {
        this._knockoutSubscriptions.push(observable.subscribe(callback));
    };
    UiElement.prototype.dispatchEvent = function (type) {
        if (this.element) {
            return this.element.dispatchEvent(new Event(type));
        }
        return false;
    };
    UiElement.prototype.addEventListener = function (type, listener, options) {
        if (!this._eventListeners) {
            this._eventListeners = new Array();
        }
        this._eventListeners.push(new eventListener_1.EventListener(type, listener, options));
    };
    UiElement.prototype.getChildren = function () {
        return this._children;
    };
    UiElement.prototype.setChildren = function (children, replace) {
        var _a;
        if (replace === void 0) { replace = true; }
        if (!children) {
            children = [];
        }
        if (replace) {
            this._children = children;
            if (this.element) {
                while (this.element.firstChild) {
                    this.element.removeChild(this.element.firstChild);
                }
            }
        }
        else {
            (_a = this._children).push.apply(_a, children);
        }
        if (this.element) {
            if (children.length > 0) {
                var documentFragment_1 = document.createDocumentFragment();
                children.forEach(function (child) {
                    documentFragment_1.appendChild(child.render());
                });
                this.element.appendChild(documentFragment_1);
            }
        }
    };
    UiElement.prototype.addChild = function (newChild, referenceChild) {
        if (referenceChild) {
            var index = this._children.indexOf(referenceChild);
            if (index > -1) {
                this._children.splice(index, 0, newChild);
                if (this.element && referenceChild.getElement()) {
                    this.element.insertBefore(newChild.render(), referenceChild.getElement());
                }
            }
        }
        else {
            this._children.push(newChild);
            if (this.element) {
                this.element.appendChild(newChild.render());
            }
        }
    };
    UiElement.prototype.removeChild = function (element) {
        var index = this._children.indexOf(element);
        if (index > -1) {
            this._children.splice(index, 1);
            if (this.element) {
                this.element.removeChild(this.element.children[index]);
            }
        }
    };
    UiElement.prototype.empty = function () {
        this._children = [];
        if (this.element) {
            while (this.element.firstChild) {
                this.element.removeChild(this.element.firstChild);
            }
        }
    };
    return UiElement;
}());
exports.UiElement = UiElement;
//# sourceMappingURL=uiElement.js.map